#include "ScalarField.h"
#include "fastmarch.hpp"
#include <math.h>

#define CREATE_DIM1  (new float[(a_NumCells+2)*(a_NumCells+2)])
#define CREATE_DIM2  (new float[(CopyField->m_NumCells+2)*(CopyField->m_NumCells+2)])
#define CREATE_DIM   (new float[(m_NumCells+2)*(m_NumCells+2)])
#define IX_DIM(i, j) ((i)+(m_NumCells+2)*(j))
#define ITER_DIM     for(int j = 1; j <= m_NumCells; j++){ for(int i = 1; i <= m_NumCells; i++){
#define ENDITER_DIM  }}

ScalarField::ScalarField(int a_NumCells, float a_Viscosity, float a_Dt)
: m_NumCells(a_NumCells), m_Field(CREATE_DIM1),new_Field(CREATE_DIM1)
, m_Viscosity(a_Viscosity), m_Dt(a_Dt),
dx(1),O(Vec2f(m_NumCells/4,2)),
dim(2),g(Vec2f(0,2))
{
	for(int i = 0; i < (m_NumCells+2)*(m_NumCells+2); i++){
		m_Field[i] = 0.01f;
		new_Field[i]=0.01f;
	}
	ITER_DIM
		if(//i<m_NumCells/2 && i>m_NumCells/6 && j>m_NumCells/10){
			j<m_NumCells/10){
			m_Field[IX_DIM(i,j)]=-1;
			new_Field[IX_DIM(i,j)]=-1;
		}else{
			m_Field[IX_DIM(i,j)]=1;
			new_Field[IX_DIM(i,j)]=1;
		}
	ENDITER_DIM
}

ScalarField::ScalarField(ScalarField *CopyField)
: m_NumCells(CopyField->m_NumCells), m_Field(CREATE_DIM2),new_Field(CREATE_DIM2),
 m_Viscosity(CopyField->m_Viscosity), m_Dt(CopyField->m_Dt),
 dim(2),g(Vec2f(0,0.1))
{
	for(int i = 0; i < (m_NumCells+2)*(m_NumCells+2); i++)
		m_Field[i] = (*CopyField)[i];
}

ScalarField::~ScalarField(void)
{
	delete[] m_Field;
	delete[] new_Field;
}
//static int step=0;
void
ScalarField::TimeStep(ScalarField *a_SrcField, VectorField *VelocityField)
{
	AddField(a_SrcField);

	//print_field(m_Field,m_NumCells);
	march(m_Field,new_Field, VelocityField->m_Field,m_NumCells);
 //   print_vf(VelocityField->m_Field,m_NumCells);

	float * tmp;
	    tmp=new_Field;
	    new_Field=m_Field;
	    m_Field=tmp;
//	print_field(m_Field,m_NumCells);
	Advection(VelocityField);
//	print_field(m_Field,m_NumCells);
	ITER_DIM
	if(//m_Field[IX_DIM(i,j)]<0 &&
        j>1){
		(*VelocityField)[IX_DIM(i,j)]-=g;
	}
	ENDITER_DIM
}

void
ScalarField::Advection(VectorField * vf)
{

    for(int i=1;i<=m_NumCells;i++){
		m_Field[IX_DIM(i,0)]=m_Field[IX_DIM(i,1)];
		m_Field[IX_DIM(i,m_NumCells+1)]=1;//m_Field[IX_DIM(i,m_NumCells)];
		m_Field[IX_DIM(0,i)]=m_Field[IX_DIM(1,i)];
		m_Field[IX_DIM(m_NumCells+1,i)]=m_Field[IX_DIM(m_NumCells,i)];
	}
	m_Field[IX_DIM(0,0)]=m_Field[IX_DIM(1,1)];
	m_Field[IX_DIM(0,m_NumCells+1)]=m_Field[IX_DIM(1,m_NumCells)];
	m_Field[IX_DIM(m_NumCells+1,0)]=m_Field[IX_DIM(m_NumCells,1)];
	m_Field[IX_DIM(m_NumCells+1,m_NumCells+1)]=m_Field[IX_DIM(m_NumCells,m_NumCells+1)];

    ITER_DIM
    Vec2f v;
    v[0]=( (*vf)[IX_DIM(i,j)][0] + (*vf)[IX_DIM(i+1,j)][0] )*0.5;
    v[1]=( (*vf)[IX_DIM(i,j)][1] + (*vf)[IX_DIM(i,j+1)][1] )*0.5;

    v=Vec2f( (0.5+i)*dx , (0.5+j)*dx ) - m_Dt*v;
    new_Field[IX_DIM(i,j)]=interp(v);
    ENDITER_DIM

    float * tmp;
    tmp=new_Field;
    new_Field=m_Field;
    m_Field=tmp;
}

float ScalarField::interp(Vec2f & pos)
{
    float f=0.f;
    //path clip
    for(int i=0;i<dim;i++){
        if(pos[i]<dx){
            pos[i]=dx;
        }
        if(pos[i]>(1+m_NumCells)*dx){
            pos[i]=(1+m_NumCells)*dx;
        }
    }

    int x=(int)(pos[0]/dx - 0.5);
    int y=(int)(pos[1]/dx - 0.5);
    float alpha=pos[0]/dx-(0.5+x);
    float beta=pos[1]/dx-(0.5+y);
    f=(1-alpha)*(1-beta)*m_Field[IX_DIM(x,y)]+
        alpha * (1-beta)*m_Field[IX_DIM(x+1,y)]+
      (1-alpha) * beta * m_Field[IX_DIM(x,y+1)]+
        alpha * beta * m_Field[IX_DIM(x+1,y+1)];
    return f;
}

void
ScalarField::AddField(ScalarField *a_SrcField)
{
	for(int i=0; i<((m_NumCells+2)*(m_NumCells+2)); i++) {
	    if(((*a_SrcField)[i])>0.01){

		m_Field[i] =- ((*a_SrcField)[i]);
	    }
	}


}

