#include "VectorField.h"
#include "ScalarField.h"
#define CREATE_DIM1 (new Vec2f[(a_NumCells+2)*(a_NumCells+2)])
#define CREATE_DIM2 (new Vec2f[(CopyField->m_NumCells+2)*(CopyField->m_NumCells+2)])
#define CREATE_DIM  (new Vec2f[(m_NumCells+2)*(m_NumCells+2)])
#define IX_DIM(i,j) ((i)+(m_NumCells+2)*(j))
#define MAT_IDX(i,j) (i-1+m_NumCells*(j-1))
#define ITER_DIM    for(int j = 1; j <= m_NumCells; j++){ for(int i = 1; i <= m_NumCells; i++){
#define ENDITER_DIM }}
void test_linsolve();

extern int neibor_dir[4][2];
void print_field(float*field, int m_NumCells);
void print_vf(Vec2f * vf, int m_NumCells);

VectorField::VectorField(int a_NumCells, float a_Viscosity, float a_Dt)
: m_NumCells(a_NumCells), m_Field(CREATE_DIM1), new_Field(CREATE_DIM1),
m_Viscosity(a_Viscosity), m_Dt(a_Dt),
dx(1),O(Vec2f(0,0)),dim(2),
vortexConfine(false),
pressure(NULL),
imat(ImpMat(m_NumCells,this)),
curl(new float[(a_NumCells+2)*(a_NumCells+2)])
{

    //test_linsolve();
	for(int i = 0; i < (m_NumCells+2)*(m_NumCells+2); i++){
		m_Field[i][0] =0.f; m_Field[i][1] = 0.f;
        new_Field[i][0] = new_Field[i][1] = 0.f;
        curl[i]=0.0f;
	}
}

VectorField::VectorField(VectorField *CopyField)
: m_NumCells(CopyField->m_NumCells), m_Field(CREATE_DIM2), new_Field(CREATE_DIM2),
m_Viscosity(CopyField->m_Viscosity), m_Dt(CopyField->m_Dt),
dim(2),vortexConfine(true),
pressure(NULL),
imat(ImpMat(m_NumCells,this)),
curl(new float[(m_NumCells+2)*(m_NumCells+2)])
{
	for(int i = 0; i < (m_NumCells+2)*(m_NumCells+2); i++){
		m_Field[i][0] = (*CopyField)[i][0];
		m_Field[i][1] = (*CopyField)[i][1];
		new_Field[i][0] = CopyField->new_Field[i][0];
		new_Field[i][1] = CopyField->new_Field[i][1];
			curl[i]=0.0f;
	}

}

VectorField::~VectorField(void)
{
	delete m_Field;
	delete new_Field;
	if(pressure){
        delete [] pressure;
	}
}

void
VectorField::TimeStep(VectorField *a_SrcField, VectorField *VelocityField)
{
	AddField(a_SrcField);
	Advection(VelocityField);
	if(vortexConfine){
       VorticityConfinement();
    }
	Projection();
}


void
VectorField::AddField(VectorField *a_SrcField)
{
	for(int i=0; i<((m_NumCells+2)*(m_NumCells+2)); i++) {
		m_Field[i][0] += m_Dt*((*a_SrcField)[i][0]);
		m_Field[i][1] += m_Dt*((*a_SrcField)[i][1]);
	}
}

float VectorField::interp(Vec2f & pos, int d)
{
    float v;
    //path clip
    for(int i=0;i<dim;i++){
        if(pos[i]<dx){
            pos[i]=dx;
        }
        if(pos[i]>(m_NumCells)*dx){
            pos[i]=(m_NumCells)*dx;
        }
    }

    int i,j;
    float alpha, beta;
    //interp vx
    if(d==1){
        i=(int)(pos[0]/dx - 0.5);
        j=(int)(pos[1]/dx );
    }else if (d==0){
        i=(int)(pos[0]/dx);
        j=(int)(pos[1]/dx - 0.5);
    }
    else{
        return 0.0;
    }

    alpha = pos[0]/dx - i;
    beta  = pos[1]/dx - j;

    if(d==1){
        alpha-=0.5f;
    }else if (d==0){
        beta-=0.5f;
    }

    v = (1-alpha)*(1-beta)*m_Field[IX_DIM(i,j)][d]
        +alpha*(1-beta)*m_Field[IX_DIM(i+1,j)][d]
        +(1-alpha)*beta * m_Field[IX_DIM(i,j+1)][d]
        +alpha*beta*m_Field[IX_DIM(i+1,j+1)][d];

    return v;
}

void VectorField::Advection(VectorField *vf)
{
    Vec2f v;

    ITER_DIM
    v=(*vf)[IX_DIM(i,j)];
    v[1]= 0.25*(v[1]+(*vf)[IX_DIM(i,j+1)][1]
                +(*vf)[IX_DIM(i-1,j)][1]
                +(*vf)[IX_DIM(i-1,j+1)][1]);
    v = Vec2f(  i * dx,(0.5+j)*dx ) - m_Dt * v;
    new_Field[IX_DIM(i,j)][0]=interp(v,0);

    v=(*vf)[IX_DIM(i,j)];
    v[0]= 0.25*(v[0]+(*vf)[IX_DIM(i+1,j)][0]
               +(*vf)[IX_DIM(i,j-1)][0]
               +(*vf)[IX_DIM(i+1,j-1)][0]);
    v = Vec2f(  (i+0.5) * dx,j*dx ) - m_Dt * v;
    new_Field[IX_DIM(i,j)][1]=interp(v,1);
    ENDITER_DIM

    Vec2f * tmp=new_Field;
    new_Field=m_Field;
    m_Field=tmp;
}

bool VectorField::hasNegativeNeibor(int i, int j)
{
    for (int k=0;k<4;k++){
        int ni=i+neibor_dir[k][0];
        int nj=j+neibor_dir[k][1];
        if((*scalar)[IX_DIM(ni,nj)]<0){
            return true;
        }
    }
    return false;
}

void print_mat(double * mat, int m_NumCells)
{
    ITER_DIM
    printf("%lf ",mat[MAT_IDX(i,j)]);
    }
    printf("\n");
    }
    printf("\n");

}

//int * matidx;
//std::vector < int > cellIdx;

void VectorField::Projection()
{
    double * x = new double[(m_NumCells)*(m_NumCells)];
    double * b = new double[(m_NumCells)*(m_NumCells)];
    if(!pressure){
        pressure=new float[(m_NumCells+2)*(2+m_NumCells)];
    }
    int steps=100;

//    cellIdx.clear();
 //   matidx=new int[(m_NumCells+2)*(2+m_NumCells)];

    ITER_DIM
        if(//hasNegativeNeibor(i,j)||
            (*scalar)[IX_DIM(i,j)]<0){
                b[MAT_IDX(i,j)]=0;
            b[MAT_IDX(i,j)]=m_Field[IX_DIM(i+1,j)][0]-m_Field[IX_DIM(i,j)][0]
                +m_Field[IX_DIM(i,j+1)][1]-m_Field[IX_DIM(i,j)][1];
          /*   if((*scalar)[IX_DIM(i+1,j)]<0){
                b[MAT_IDX(i,j)]+=m_Field[IX_DIM(i+1,j)][0];
             }
             if((*scalar)[IX_DIM(i-1,j)]<0){
                b[MAT_IDX(i,j)]-=m_Field[IX_DIM(i,  j)][0];
             }
             if((*scalar)[IX_DIM(i,j+1)]<0){
                b[MAT_IDX(i,j)]+=m_Field[IX_DIM(i,j+1)][1];
             }
             if((*scalar)[IX_DIM(i,j-1)]<0){
                b[MAT_IDX(i,j)]-=m_Field[IX_DIM(i,j)][1];
             }*/
        }else{
            b[MAT_IDX(i,j)]=0;
        }
        x[MAT_IDX(i,j)]=0;
    ENDITER_DIM
  //  b[0]=0;
    ConjGrad(m_NumCells*m_NumCells,&imat, x, b, 0.000001, &steps);
    //print_mat(x,m_NumCells);
    printf("%d\n",steps);

//    print_vf(m_Field, m_NumCells);

    ITER_DIM
    if(hasNegativeNeibor(i,j)||
       (*scalar)[IX_DIM(i,j)]<0){
        if(i>1){
            m_Field[IX_DIM(i,j)][0]-=
                x[MAT_IDX(i,j)]-x[MAT_IDX(i-1,j)];
        }else{
            m_Field[IX_DIM(i,j)][0]=0;
        }
        if(j>1){
            m_Field[IX_DIM(i,j)][1]-=
                x[MAT_IDX(i,j)]-x[MAT_IDX(i,j-1)];
        }else{
            m_Field[IX_DIM(i,j)][1]=0;
        }
    }
        pressure[IX_DIM(i,j)]=x[MAT_IDX(i,j)];
    ENDITER_DIM

 //   printf("after projection:\n");
 //   print_vf(m_Field, m_NumCells);

 //   printf("pressurefield:\n");
 //   print_field(pressure,m_NumCells);
    delete x;
    delete b;
}

ImpMat::ImpMat(int num,VectorField* vecf):
m_NumCells(num),vf(vecf)
{}

/*
void test_linsolve()
{
    int size=2;
    double *b=new double [size*size];
    double * x=new double [size*size];
    int step=20;
    ImpMat mat(2,this);

    for(int i=0;i<size;i++){
        for(int j=0;j<size;j++){
            b[i+j*size]=i+j*size;
            printf("%f\t",b[i+j*size]);
        }
        printf("\n");
    }
    b[0]=0;
    ConjGrad(size*size , &mat, x, b,0.0001,&step);
    printf("%d\n",step);
    for(int i=0;i<size;i++){
        for(int j=0;j<size;j++){

            printf("%f\t",x[i+j*size]);
        }
        printf("\n");
    }

    delete b;
    delete x;
}
*/
void ImpMat::matVecMult(double x[], double r[])
{
    int middle;
    ITER_DIM
    if(//vf->hasNegativeNeibor(i,j)||
        (*(vf->scalar))[IX_DIM(i,j)]<0){

        middle=0;
        r[MAT_IDX(i,j)]=0;
        if(i>1){
            middle--;
            if((*(vf->scalar))[IX_DIM(i-1,j)]<0){
                r[MAT_IDX(i,j)]+=x[MAT_IDX(i-1,j)];
            }
        }

        if(i<m_NumCells){
            middle--;
            if((*(vf->scalar))[IX_DIM(i+1,j)]<0){
                r[MAT_IDX(i,j)]+=x[MAT_IDX(i+1,j)];
            }
        }

        if(j>1){
            middle--;
            if((*(vf->scalar))[IX_DIM(i,j-1)]<0){
                r[MAT_IDX(i,j)]+=x[MAT_IDX(i,j-1)];
            }
        }

        if(j<m_NumCells){
            middle--;
            if((*(vf->scalar))[IX_DIM(i,j+1)]<0){
                r[MAT_IDX(i,j)]+=x[MAT_IDX(i,j+1)];
            }
        }
        r[MAT_IDX(i,j)]+=middle*x[MAT_IDX(i,j)];
    }

    else{
        r[MAT_IDX(i,j)]=x[MAT_IDX(i,j)];
    }

    ENDITER_DIM
 /*       printf("==========\n");
    for(int i=0;i<m_NumCells;i++){
        for(int j=0;j<m_NumCells;j++){

            printf("%f\t",x[i+j*m_NumCells]);
        }
        printf("\n");
    }
        printf("\n");
    for(int i=0;i<m_NumCells;i++){
        for(int j=0;j<m_NumCells;j++){

            printf("%f\t",r[i+j*m_NumCells]);
        }
        printf("\n");
    }
            printf("\n");
        printf("==========\n");
*/
  //  r[0]=x[0];
}

/**@brief currently only written for 2d
*/
float epsilon=0.1f;
void VectorField::VorticityConfinement()
{
    //curl
    double omega, duydx, duxdy;
    Vec2f omegaGrad;
    ITER_DIM
        duydx=0.25*(m_Field[IX_DIM(i+1,j)][1]-
            m_Field[IX_DIM(i-1,j)][1] +
            m_Field[IX_DIM(i+1,j+1)][1]-
            m_Field[IX_DIM(i-1,j+1)][1]);
        duxdy=0.25*(m_Field[IX_DIM(i,j+1)][0]-
            m_Field[IX_DIM(i,j-1)][0] +
            m_Field[IX_DIM(i+1,j+1)][0]-
            m_Field[IX_DIM(i+1,j-1)][0]);
        omega=duydx-duxdy;
        curl[IX_DIM(i,j)]=abs(omega);
    ENDITER_DIM

    ITER_DIM
        omegaGrad[0]=0.25*(curl[IX_DIM(i+1,j)]-
            curl[IX_DIM(i-1,j)] +
            curl[IX_DIM(i+1,j-1)]-
            curl[IX_DIM(i-1,j-1)]);
        omegaGrad[1]=0.25*(curl[IX_DIM(i,j+1)]-
            curl[IX_DIM(i,j-1)] +
            curl[IX_DIM(i-1,j+1)]-
            curl[IX_DIM(i-1,j-1)]);
        unitize(omegaGrad);

        duydx=0.25*(m_Field[IX_DIM(i+1,j)][1]-
            m_Field[IX_DIM(i-1,j)][1] +
            m_Field[IX_DIM(i+1,j+1)][1]-
            m_Field[IX_DIM(i-1,j+1)][1]);
        duxdy=0.25*(m_Field[IX_DIM(i,j+1)][0]-
            m_Field[IX_DIM(i,j-1)][0] +
            m_Field[IX_DIM(i+1,j+1)][0]-
            m_Field[IX_DIM(i+1,j-1)][0]);
        omega=duydx-duxdy;

        m_Field[IX_DIM(i,j)][0]+=epsilon *
            (omega*omegaGrad[1]);
        m_Field[IX_DIM(i,j)][1]+=epsilon *
            (-omega*omegaGrad[0]);
    ENDITER_DIM
}

