#pragma once
#include <gfx/vec2.h>
#include "linearSolver.h"

class ScalarField;
class VectorField;
class ImpMat:public implicitMatrix
{
public:
    void matVecMult(double x[], double r[]);
    /**@brief width of the grid*/
    int m_NumCells;
    ImpMat(int num,VectorField * vf);
    VectorField * vf;
};

class VectorField
{
public:
	VectorField(int a_NumCells, float a_Viscosity, float a_Dt);
	VectorField(VectorField *CopyField);
	virtual ~VectorField(void);

	void	TimeStep(VectorField *a_SrcField, VectorField *VelocityField);
	void	AddField(VectorField *a_SrcField);
	void	Advection(VectorField *vf);
	void    VorticityConfinement();
	void    Projection();
	/**@brief interpolate velocity at position x,y
	x y are in the field's coordinates
	@param d which dimension
	*/
    float   interp(Vec2f & pos,int d);

	Vec2f	&operator[](int i)       { return m_Field[i]; }

	int		m_NumCells;
	Vec2f	*m_Field;
	Vec2f	*new_Field;

	float	m_Viscosity;
	float	m_Dt;
	float   dx;
	/**@brief origin of the field
        only used for rendering
        when computing origin is 0,0)
	*/
	Vec2f   O;
	int dim;
	bool vortexConfine;

    ScalarField * scalar;
    bool hasNegativeNeibor(int i, int j);
    float   * pressure;
private:
    ImpMat imat;
    float * curl;

};


