#include "fastmarch.hpp"

#define IX_DIM(i, j) ((i)+(m_NumCells+2)*(j))
#define ITER_DIM     for(int j = 1; j <= m_NumCells; j++){ for(int i = 1; i <= m_NumCells; i++){
#define ITER_ALL     for(int j = 0; j <= (1+m_NumCells); j++){ for(int i = 0; i <= (1+m_NumCells); i++){
#define ENDITER_DIM  }}

#define SOLVE_D(x,y) ((x)*(y)/sqrt((x)*(x)+(y)*(y)))
#define SOLVE_QUAD(x,y) (((x)+(y)+sqrt(2-(x-y)*(x-y)))*0.5)
#define IS_INSIDE(x,y) ((x)>0 && (x)<(m_NumCells+1) && (y)>0 && (y)<(m_NumCells+1))
/**@brief practically impossible to simulate on
such a large grid*/
#define POS_INFTY 100000
#define EPSILON 0.00001
#define MIN_FLOAT(x,y) ( (x)<(y) ? (x):(y) )
#define MIN_IDX(x,y) ( (x)<(y) ? 0 : 1 )
#define MAX_FLOAT(x,y) ( (x)<(y) ? (y):(x) )

/******
Global Variables
**********************/
int neibor_dir[4][2]={{0,1},{1,0},{0,-1},{-1,0}};
vector<Status>state;
Vec2f * vf;
//is velocity being extended
bool extendv;

void print_state(int m_NumCells);
Vec2f weightAverage(float s, Vec2f vs, float t, Vec2f vt);
/**********
implementations
**********/
float INTERSECT(float x,float y)
{
    if(x>EPSILON || y>EPSILON){
        return x/(x-y);
    }
    return 0.5f;
}

void update_cell(int i, int j,float * dst,int m_NumCells)
{
	//x plus, x minus, y plus, y minus
	float xp,xm,yp,ym;
	Vec2f vx[2], vy[2];
	int xcnt=0,ycnt=0;
	if(i<m_NumCells &&
		(state[IX_DIM(i+1,j)]==ACCEPTED)){
		xp=dst[IX_DIM(i+1,j)];
		xcnt=1;
		vx[0]=vf[IX_DIM(i+1,j)];
	}else{xp=POS_INFTY;}
	if(i>1 &&
		(state[IX_DIM(i-1,j)]==ACCEPTED)){
		xm=dst[IX_DIM(i-1,j)];
		xcnt=1;
		vx[1]=vf[IX_DIM(i-1,j)];
	}else{xm=POS_INFTY;}
	if(j<m_NumCells
		&&(state[IX_DIM(i,j+1)]==ACCEPTED)){
		yp=dst[IX_DIM(i,j+1)];
		ycnt=1;
		vy[0]=vf[IX_DIM(i,j+1)];
	}else{yp=POS_INFTY;}
	if(j>1
		&&(state[IX_DIM(i,j-1)]==ACCEPTED)){
		ym=dst[IX_DIM(i,j-1)];
		ycnt=1;
        vy[1]=vf[IX_DIM(i,j-1)];
	}else{ym=POS_INFTY;}

	xp=MIN_FLOAT( xm, xp);
	vx[0]=vx[MIN_IDX(xp,xm)];
	vy[0]=vy[MIN_IDX(yp,ym)];
	yp=MIN_FLOAT( ym, yp);
	if(xcnt>0 && ycnt>0){
		dst[IX_DIM(i,j)]=SOLVE_QUAD(xp,yp);
		float weightx=abs(dst[IX_DIM(i,j)]-xp);
		float weighty=abs(dst[IX_DIM(i,j)]-yp);
        if(extendv){
            vf[IX_DIM(i,j)]=(weightx*vx[0]+weighty*vy[0])/(weightx+weighty);
        }
	}else if(xcnt>0){
		dst[IX_DIM(i,j)]=xp+1;
		if(extendv){
		    vf[IX_DIM(i,j)]=vx[0];
		}
	}else{
		dst[IX_DIM(i,j)]=yp+1;
		if(extendv){
		    vf[IX_DIM(i,j)]=vy[0];
		}
	}

}
/**@brief when update neibors of TRIAL,
never remove them from the priority queue,
just add duplicated cells into queue.
During processing, if a cell is already
processed discard it.
*/
void redist(float * field, float * dst, int m_NumCells)
{

	state.resize((m_NumCells+2)*(m_NumCells+2));
	mycmp cmp(dst);
	PQ pq(cmp);
	field_init(field,dst , m_NumCells,pq);
//	print_state(m_NumCells);
	while(!pq.empty()){
		intfloat cell=pq.top();
		pq.pop();
		int i=cell.first%(m_NumCells+2);
		int j=cell.first/(m_NumCells+2);
		if(state[IX_DIM(i,j)]==ACCEPTED){
			continue;
		}
		state[IX_DIM(i,j)]=ACCEPTED;
		update_cell(i,j,dst,m_NumCells);
		for(int k=0;k<4;k++){
			int ni=i+neibor_dir[k][0];
			int nj=j+neibor_dir[k][1];
			if((!IS_INSIDE(ni,nj)) ||
                (state[IX_DIM(ni,nj)]==ACCEPTED)){
				continue;
			}
			state[IX_DIM(ni,nj)]=CLOSE;
			update_cell(ni,nj,dst,m_NumCells);
			pq.push(make_pair(IX_DIM(ni,nj),dst[IX_DIM(ni,nj)]));
		}
	}
}

void march(float * field, float * dst, Vec2f *vecf, int m_NumCells)
{
    vf=vecf;
    //print_vf(m_NumCells);
    //print_field(field, m_NumCells);
    extendv=true;
	//boundaries
	for(int i=0;i<=m_NumCells+1;i++){
		field[IX_DIM(i,0)]=field[IX_DIM(i,1)];
		field[IX_DIM(m_NumCells+1,i)]=field[IX_DIM(m_NumCells,i)];
		field[IX_DIM(0,i)]=field[IX_DIM(1,i)];
		field[IX_DIM(i,m_NumCells+1)]=1.0f;
	}

	redist(field, dst, m_NumCells);
	ITER_ALL
		field[IX_DIM(i,j)]=-field[IX_DIM(i,j)];
	ENDITER_DIM
	ITER_ALL
		dst[IX_DIM(i,j)]=-dst[IX_DIM(i,j)];
	ENDITER_DIM

//print_vf(m_NumCells);


	extendv=false;
	redist(field, dst, m_NumCells);
	ITER_DIM
		dst[IX_DIM(i,j)]=-dst[IX_DIM(i,j)];
	ENDITER_DIM
}

Status init_point(int i, int j, float*src,float*dst,int m_NumCells)
{
	//index of cell outside the surface
	int neibor_cnt=0, out_idx=0;
	float s[4];
	Vec2f v[4];
	//indices of neibors inside surface
	char idx[4];
	//neibor indicies
	int ni,nj;
	int minidx=0;
	float min=1000000;
	for(int k=0;k<4;k++){
		ni=i+neibor_dir[k][0];
		nj=j+neibor_dir[k][1];
		if(IS_INSIDE(ni,nj) && src[IX_DIM(ni,nj)]<0){
			s[neibor_cnt]=INTERSECT(src[IX_DIM(i,j)],
					        src[IX_DIM(ni,nj)]);
            v[neibor_cnt]=vf[IX_DIM(ni, nj)];
			idx[neibor_cnt]=k;
			neibor_cnt++;

			if(src[IX_DIM(i,j)]<min){
			    min=src[IX_DIM(i,j)];
			    minidx=k;
			}

		}else{
			out_idx=k;
		}
	}
	float t=0;

	switch(neibor_cnt){
	case 0:
		dst[IX_DIM(i,j)]=POS_INFTY;
		if(extendv){
        //    vf[IX_DIM(i,j)]=Vec2f(0.0 , 0.0);
		}
		return FAR;
	case 1:
		dst[IX_DIM(i,j)]=s[0];
		if(extendv ){
         //   vf[IX_DIM(i,j)]=v[0];
		}
		break;
	case 2:
		/*4d
		-------

		-------
		*/
		if(idx[1]-idx[0]==2){
		    int minIdx=MIN_IDX(s[0],s[1]);
			dst[IX_DIM(i,j)]=s[minIdx];
			if(extendv ){
			//    vf[IX_DIM(i,j)]=v[minIdx];
			}
		}
		/*4b
		     |
		____/
		*/
		else{
		    dst[IX_DIM(i,j)]=SOLVE_D(s[0],s[1]);
			if(extendv){
			//    vf[IX_DIM(i,j)]=weightAverage(s[0], v[0], s[1],v[1]);
			}
		}
		break;
	case 3:
		{int k;
		Vec2f vt;
		for(k=0;k<3;k++){
			if(idx[k]==(out_idx+2)%4){
				t=s[k];
				vt=v[k];
				if(k<2){
					s[k]=s[2];
					v[k]=v[2];
				}
				break;
			}
		}
		int minIdx=MIN_IDX(s[0],s[1]);
		dst[IX_DIM(i,j)]=SOLVE_D(s[minIdx],t);
        if(extendv ){
         //  vf[IX_DIM(i,j)]=weightAverage(s[minIdx],v[minIdx], t, vt);
        }
		break;}
	case 4:{
	    int minIdx=2 * MIN_IDX(s[0],s[2]);
        float min_s=s[minIdx];
        Vec2f vs=v[minIdx];

        minIdx=1+2 * MIN_IDX(s[1],s[3]);
        float min_t=s[minIdx];
        Vec2f vt=v[minIdx];

		dst[IX_DIM(i,j)]=SOLVE_D(min_s,min_t);
        if(extendv && dst[IX_DIM(i,j)]>0.5){
         //  vf[IX_DIM(i,j)]=weightAverage(min_s, vs, min_t, vt);
        }
		break;
        }
	}
    if(extendv && src[IX_DIM(i,j-1)]>0){
        ni=i+neibor_dir[minidx][0];
        nj=j+neibor_dir[minidx][1];
        vf[IX_DIM(i,j)][1]=vf[IX_DIM(ni,nj)][1];
	}
	if(extendv && src[IX_DIM(i-1,j)]>0){
        ni=i+neibor_dir[minidx][0];
        nj=j+neibor_dir[minidx][1];
        vf[IX_DIM(i,j)][0]=vf[IX_DIM(ni,nj)][0];
	}

	return ACCEPTED;
}

void field_init(float * src,float*dst, int m_NumCells,PQ& pq)
{
	Status status;
	vector<pair<int, int> >accepted;

		//boundaries
	for(int i=0;i<=m_NumCells+1;i++){
		state[IX_DIM(i,0)]=ACCEPTED;
		state[IX_DIM(i,m_NumCells+1)]=ACCEPTED;
		state[IX_DIM(m_NumCells+1,i)]=ACCEPTED;
		state[IX_DIM(0,i)]=ACCEPTED;
	}

	ITER_DIM
	if(src[IX_DIM(i,j)]>=0){
		status=init_point(i,j,src,dst,m_NumCells);
		state[IX_DIM(i,j)]=status;
		if(status==ACCEPTED){
			accepted.push_back(make_pair(i,j));
		}
	}else{
	//	dst[IX_DIM(i,j)]=src[IX_DIM(i,j)];
		state[IX_DIM(i,j)]=ACCEPTED;
	}
	ENDITER_DIM
/*set status of boundary cells to ACCEPTED
  to avoid adding them into queue
*/

//neibor indicies
	int ni,nj;
	for(int i=0;i<accepted.size();i++){
		for(int k=0;k<4;k++){
			ni=neibor_dir[k][0]+accepted[i].first;
			nj=neibor_dir[k][1]+accepted[i].second;
			if(!IS_INSIDE(ni,nj)){
                continue;
			}
			if(state[IX_DIM(ni,nj)]==FAR){
				state[IX_DIM(ni,nj)]=CLOSE;
				update_cell(ni,nj,dst,m_NumCells);
				pq.push(make_pair(IX_DIM(ni,nj),dst[IX_DIM(ni,nj)]));
			}
		}
	}
//	printf("init field\n");
//	print_field( dst, m_NumCells);
}

void print_field(float * field, int m_NumCells)
{
	printf("\n=========\n");
	ITER_DIM
	printf("%.3f ",field[IX_DIM(i,m_NumCells-j+1)]);
	}
	printf("\n");
	}
	printf("\n=========\n");
}

void print_state(int m_NumCells)
{
	printf("\n=========\n");
	ITER_DIM
	printf("%d ",state[IX_DIM(i,m_NumCells-j+1)]);
	}
	printf("\n");
	}
	printf("\n=========\n");
}

void print_vf(Vec2f * vf, int m_NumCells)
{
    printf("\n=========\n");

	ITER_ALL


	printf("(%3.2lf,%3.3lf) ",vf[IX_DIM(i,m_NumCells-j+1)][0],vf[IX_DIM(i,m_NumCells-j+1)][1]);

	}
	printf("\n");
	}

	printf("\n=========\n");
}

Vec2f weightAverage(float s, Vec2f vs, float t, Vec2f vt)
{
    s=s*s;
    t=t*t;
    if(s<EPSILON){
        s=1;
    }
    if(t<EPSILON){
        t=1;
    }
    s=1/s;
    t=1/t;
    return (s*vs+t*vt)/(s+t);
}
